import os
import tempfile

import numpy as np

from sparrow_datums import BoxTracking

from .reader import ChunkStreamReader
from .writer import ChunkStreamWriter


def test_chunk_stream_reader():
    with tempfile.TemporaryDirectory() as dir:
        manifest_path = os.path.join(dir, "stream.jsonl")
        writer = ChunkStreamWriter(manifest_path, BoxTracking, fps=1)
        a = BoxTracking(np.ones((2, 2, 4)), fps=1, start_time=0)
        b = BoxTracking(np.ones((2, 2, 4)), fps=1, start_time=a.duration)
        writer.add_chunk(a)
        writer.add_chunk(b)
        writer.close()
        reader = ChunkStreamReader(manifest_path, BoxTracking)
        assert reader.header.fps == 1
        assert len(reader) == 2
        assert isinstance(reader[1], BoxTracking)



from __future__ import annotations

from pathlib import Path
from typing import Optional

from ..chunk import T
from ..types import PType
from .types import ChunkPath, Footer, Header


class ChunkStreamWriter:
    """A class for writing chunk streams to disk."""

    def __init__(
        self,
        manifest_path: str | Path,
        chunk_type: type[T],
        fps: Optional[float] = None,
        start_time: float = 0,
        ptype: Optional[PType] = None,
        image_width: Optional[int] = None,
        image_height: Optional[int] = None,
    ) -> None:
        self.manifest_path = Path(manifest_path)
        self.chunk_type = chunk_type
        self.header = Header(
            classname=chunk_type.__name__,
            ptype=ptype.name if ptype else None,
            image_width=image_width,
            image_height=image_height,
            fps=fps,
            start_time=start_time,
        )
        self.chunk_paths: list[ChunkPath] = []
        self.footer = Footer(is_done=False)
        self.next_start_time = start_time
        self.write_manifest()

    def __enter__(self) -> "ChunkStreamWriter":
        """Enter context manager."""
        return self

    def __exit__(self, *args, **kwargs) -> None:
        """Exit context manager."""
        self.close()

    def add_chunk(self, chunk: T) -> None:
        """Add chunk to stream and re-write the manifest."""
        # Fill missing header values
        if self.header.ptype is None:
            self.header.ptype = chunk.ptype.name
        if self.header.image_width is None:
            self.header.image_width = chunk._image_width
        if self.header.image_height is None:
            self.header.image_height = chunk._image_height
        if self.header.fps is None:
            self.header.fps = chunk._fps
        # Validation
        if not isinstance(chunk, self.chunk_type):
            raise TypeError(
                (f"Incorrect chunk type {type(chunk)}. Expected {self.chunk_type}.")
            )
        if self.header.ptype != chunk.ptype.name:
            raise ValueError(
                f"Incorrect PType {chunk.ptype.name}. Expected {self.header.ptype}."
            )
        if self.header.image_width != chunk._image_width:
            raise ValueError(
                f"Incorrect image_width {chunk._image_width}. Expected {self.header.image_width}."
            )
        if self.header.image_height != chunk._image_height:
            raise ValueError(
                f"Incorrect image_height {chunk._image_height}. Expected {self.header.image_height}."
            )
        if self.header.fps != chunk._fps:
            raise ValueError(f"Incorrect fps {chunk._fps}. Expected {self.header.fps}.")
        if chunk.start_time != self.next_start_time:
            raise ValueError(
                f"Incorrect start time {chunk.start_time}. Expected {self.next_start_time}."
            )
        chunk_index = len(self.chunk_paths)
        path = f"{chunk_index:04d}.json.gz"
        # If duration/start_time is missing,
        # we want to fail before we write the chunk to disk.
        duration = chunk.duration
        start_time = chunk.start_time
        chunk.to_file(self.manifest_path.parent / path)
        self.chunk_paths.append(
            ChunkPath(path=path, start_time=start_time, duration=duration)
        )
        self.write_manifest()
        self.next_start_time += duration

    def write_manifest(self) -> None:
        """Write the manifest with all the chunks."""
        values = (
            [self.header.to_json()]
            + [p.to_json() for p in self.chunk_paths]
            + [self.footer.to_json()]
        )
        with open(self.manifest_path, "w") as f:
            f.write("\n".join(values))

    def close(self) -> None:
        """Close the chunk stream."""
        self.footer = Footer(is_done=True)
        self.write_manifest()



from __future__ import annotations

import json
from dataclasses import asdict, dataclass
from typing import Dict, TypeVar, Union

T = TypeVar("T", bound="StreamValue")
ValueDict = Dict[str, Union[str, int, float, bool]]


@dataclass
class StreamValue:
    """Base class for JSON lines values in manifest file."""

    def to_dict(self) -> ValueDict:
        """Serialize the value as a dict."""
        return asdict(self)

    def to_json(self) -> str:
        """Serialize the value as a JSON string."""
        return json.dumps(self.to_dict())

    @classmethod
    def from_dict(cls: type[T], value_dict: ValueDict) -> T:
        """Create a StreamValue object from a dict."""
        return cls(**value_dict)

    @classmethod
    def from_json(cls: type[T], value_str: str) -> T:
        """Create a StreamValue object from a JSON string."""
        return cls.from_dict(json.loads(value_str))


@dataclass
class Header(StreamValue):
    """
    Header information for a chunk stream manifest.

    This is roughly equivalent to chunk metadata.
    """

    classname: str
    ptype: Union[str, None]
    image_width: Union[int, None]
    image_height: Union[int, None]
    fps: float
    start_time: float


@dataclass
class ChunkPath(StreamValue):
    """Information about a chunk in the stream for the manifest."""

    path: str
    start_time: float
    duration: float


@dataclass
class Footer(StreamValue):
    """Footer information for a chunk stream manifest."""

    is_done: bool



"""Intersection over union operations."""
from typing import Union, no_type_check

import numpy as np
import numpy.typing as npt

from .boxes import Boxes
from .frame_augmented_boxes import FrameAugmentedBoxes
from .frame_boxes import FrameBoxes


def intersection(boxes_a: Boxes, boxes_b: Boxes) -> npt.NDArray[np.float64]:
    """
    Intersection areas of two sets of boxes.

    Parameters
    ----------
    boxes_a : Boxes
        A boxes array
    boxes_b : Boxes
        A boxes array

    Returns
    -------
    intersections : np.ndarray
        An ndarray of intersection areas
    """
    # Define the inner box
    x1 = np.maximum(boxes_a.x1, boxes_b.x1)
    y1 = np.maximum(boxes_a.y1, boxes_b.y1)
    x2 = np.minimum(boxes_a.x2, boxes_b.x2)
    y2 = np.minimum(boxes_a.y2, boxes_b.y2)
    result: npt.NDArray[np.float64] = np.maximum(x2 - x1, 0) * np.maximum(y2 - y1, 0)
    return result


def area(boxes: Boxes) -> npt.NDArray[np.float64]:
    """
    Areas of boxes.

    Parameters
    ----------
    boxes : Boxes
        A boxes array

    Returns
    -------
    areas : np.ndarray
        An ndarray of areas
    """
    return boxes.w * boxes.h


@no_type_check
def pairwise_iou(
    boxes_a: Union[FrameAugmentedBoxes, FrameBoxes],
    boxes_b: Union[FrameAugmentedBoxes, FrameBoxes],
) -> npt.NDArray[np.float64]:
    """
    Pairwise IoU for two sets of frame boxes.

    Parameters
    ----------
    boxes_a : FrameAugmentedBoxes or FrameBoxes
        A frame of boxes
    boxes_b : FrameAugmentedBoxes or FrameBoxes
        A frame of boxes

    Returns
    -------
    iou_scores : np.ndarray
        A (n_boxes_a, n_boxes_b) matrix of iou scores
    """
    boxes_a_parent = boxes_a.__class__.__base__
    boxes_a_view: Union[FrameAugmentedBoxes, FrameBoxes] = boxes_a.view(boxes_a_parent)
    boxes_a_expanded = boxes_a_view[:, None, :]
    boxes_b_parent = boxes_b.__class__.__base__
    boxes_b_view: Union[FrameAugmentedBoxes, FrameBoxes] = boxes_b.view(boxes_b_parent)
    boxes_b_expanded = boxes_b_view[None, :, :]
    intersections = intersection(boxes_a_expanded, boxes_b_expanded)
    areas = area(boxes_a_expanded) + area(boxes_b_expanded)
    unions = areas - intersections
    unions[unions == 0] = np.inf
    return intersections / unions



import doctest

import numpy as np
import numpy.typing as npt
import pytest

from ..types import PType
from . import augmented_boxes
from .augmented_boxes import AugmentedBoxes


def test_docstring_example():
    result = doctest.testmod(augmented_boxes)
    assert result.attempted > 0
    assert result.failed == 0


def test_augmented_boxes_with_6d_array_succeeds():
    x: npt.NDArray[np.float64] = np.ones((5, 6))
    boxes = AugmentedBoxes(x, ptype=PType.relative_tlwh)
    assert boxes.ptype == PType.relative_tlwh


def test_wrong_shape_throws_value_error():
    x: npt.NDArray[np.float64] = np.ones((5, 4))
    with pytest.raises(ValueError):
        AugmentedBoxes(x)


def test_non_integer_labels_throws_value_error():
    x: npt.NDArray[np.float64] = np.ones((5, 6))
    x[..., -1] += np.random.uniform(size=5)
    with pytest.raises(ValueError):
        AugmentedBoxes(x)


def test_invalid_scores_range_throws_value_error():
    x: npt.NDArray[np.float64] = np.ones((5, 6))
    x[..., -2] += 1
    with pytest.raises(ValueError):
        AugmentedBoxes(x)


def test_to_relative_moves_boxes_to_0_1():
    image_size = 512
    x: npt.NDArray[np.float64] = np.random.uniform(size=(10, 6))
    x[..., :4] *= image_size
    x[..., -1] = np.round(x[..., -1])
    boxes_a = AugmentedBoxes(
        x,
        PType.absolute_tlwh,
        image_width=image_size,
        image_height=image_size,
    )
    boxes_b = boxes_a.to_relative()
    assert boxes_b.is_relative
    np.testing.assert_array_less(boxes_b.array, 1.01)
    boxes_c = boxes_b.to_relative()
    assert boxes_c.is_relative


def test_to_absolute_moves_boxes_to_0_image_size():
    image_size = 512
    x = np.random.uniform(size=(10, 6))
    x[..., -1] = np.round(x[..., -1])
    boxes_a = AugmentedBoxes(
        x,
        PType.relative_tlwh,
        image_width=image_size,
        image_height=image_size,
    )
    boxes_b = boxes_a.to_absolute()
    assert boxes_b.is_absolute
    assert (boxes_b.array[..., :4] < 1).mean() < 0.1
    boxes_c = boxes_b.to_absolute()
    assert boxes_c.is_absolute


def test_to_tlbr_moves_boxes_to_tlbr():
    x: npt.NDArray[np.float64] = np.concatenate([np.ones((5, 2)), np.zeros((5, 4))], -1)
    boxes = AugmentedBoxes(x, PType.relative_tlwh).to_tlbr()
    np.testing.assert_equal(boxes.array[..., :4], 1)
    np.testing.assert_equal(boxes.array[..., 4:], 0)


def test_to_tlwh_moves_boxes_to_tlwh():
    result: npt.NDArray[np.float64] = np.concatenate(
        [np.ones((5, 2)), np.zeros((5, 2))], -1
    )
    boxes = AugmentedBoxes(np.ones((5, 6)), PType.relative_tlbr).to_tlwh()
    np.testing.assert_equal(boxes.array[..., :4], result)
    np.testing.assert_equal(boxes.array[..., 4:], 1)


def test_scores_and_labels_attributes():
    boxes_array = np.ones((6, 4))
    scores = np.random.uniform(size=6)
    labels = np.random.randint(10, size=6)
    boxes = AugmentedBoxes(
        np.concatenate([boxes_array, scores[:, None], labels[:, None]], -1)
    )
    np.testing.assert_almost_equal(scores, boxes.scores)
    np.testing.assert_almost_equal(labels, boxes.labels)


def test_multi_dimensional_label_names():
    label_names = ["a", "b", "c", "d", "e", "f"]
    boxes = np.ones((6, 4))
    scores: npt.NDArray[np.float64] = np.random.uniform(size=6)
    labels = np.arange(6)
    x: npt.NDArray[np.float64] = np.concatenate(
        [boxes, scores[:, None], labels[:, None]], -1
    )
    boxes = AugmentedBoxes(x.reshape(2, 3, 6))
    np.testing.assert_equal(
        boxes.names(label_names), np.array(label_names).reshape(2, 3)
    )



import numpy as np

from ..types import PType
from .boxes import Boxes
from .frame_augmented_boxes import FrameAugmentedBoxes
from .frame_boxes import FrameBoxes
from .iou import area, intersection, pairwise_iou
from .single_box import SingleBox


def test_intersection_calculation():
    box_a = SingleBox(np.ones(4), PType.absolute_tlwh)
    box_b = SingleBox(np.ones(4) * 2, PType.absolute_tlwh)
    assert intersection(box_a, box_a) == 1
    assert intersection(box_b, box_b) == 4
    assert intersection(box_a, box_b) == 0


def test_intersection_works_on_frame_boxes():
    boxes = FrameBoxes(np.ones((4, 4)), PType.absolute_tlwh)
    result = intersection(boxes, boxes)
    np.testing.assert_equal(result, np.ones(4))


def test_pairwise_intersection_works():
    boxes = Boxes(np.ones((5, 4)), PType.absolute_tlwh)
    result = intersection(boxes[:, None], boxes[None])
    assert result.shape == (5, 5)
    np.testing.assert_equal(result, 1)


def test_area_calculation():
    box = SingleBox(np.ones(4) * 3, PType.absolute_tlwh)
    assert area(box) == 9
    boxes = FrameBoxes(np.stack([np.arange(3)] * 4, -1), PType.absolute_tlwh)
    np.testing.assert_array_equal(area(boxes), np.array([0, 1, 4]))


def test_pairwise_iou_creates_pairwise_shape():
    boxes = FrameBoxes(np.ones((5, 4)), PType.absolute_tlwh)
    result = pairwise_iou(boxes, boxes)
    assert result.shape == (5, 5)


def test_pairwise_iou_works_for_frame_augmented_boxes():
    boxes = FrameAugmentedBoxes(np.ones((5, 6)), PType.absolute_tlwh)
    result = pairwise_iou(boxes, boxes)
    assert result.shape == (5, 5)


def test_max_iou_is_1():
    frame_a = FrameAugmentedBoxes(
        [[130.02, 582.8, 87.96, 57.14, 1.0, 2.0]],
        PType.absolute_tlwh,
        image_width=1920,
        image_height=1080,
    )
    frame_b = FrameAugmentedBoxes(
        [[1.82, 499.01, 41.07, 38.85, 1.0, 3.0]],
        PType.absolute_tlwh,
        image_width=1920,
        image_height=1080,
    )
    assert pairwise_iou(frame_a, frame_b) < 1



from typing import TypeVar

import numpy as np

from ..chunk import Chunk

T = TypeVar("T", bound="Tracking")


class Tracking(Chunk):
    """Validate that the chunk has 3 dimensions."""

    def validate(self) -> None:
        """Check shape of tracking array."""
        if self.ndim != 3:
            raise ValueError("Tracking chunks must have 3 dimensions")

    @property
    def duration(self) -> float:
        """Compute duration (in seconds) of tracking chunk."""
        return len(self) / self.fps

    def resample(self: T, new_fps: float) -> T:
        """Modify fps for a tracking chunk."""
        n_current_frames = len(self)
        duration_seconds = n_current_frames / self.fps
        n_new_frames = round(duration_seconds * new_fps)
        frame_shape = self.shape[1:]
        data = np.zeros((n_new_frames,) + frame_shape) * np.nan
        for new_idx in range(n_new_frames):
            current_idx = int((new_idx / new_fps) * self.fps)
            data[new_idx] = self.array[current_idx]
        metadata_kwargs = self.metadata_kwargs.copy()
        metadata_kwargs["fps"] = new_fps
        return self.__class__(
            data=data,
            ptype=self.ptype,
            **metadata_kwargs,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sparrow_datums',
 'sparrow_datums.boxes',
 'sparrow_datums.stream',
 'sparrow_datums.tracking']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.20.0,<2.0.0', 'rich>=10.14.0,<11.0.0', 'typer[all]>=0.4.0,<0.5.0']

setup_kwargs = {
    'name': 'sparrow-datums',
    'version': '0.8.1.dev1655837204',
    'description': 'A Python package for data structures',
    'long_description': '# sparrow-datums\n\n<div align="center">\n\nA Python package for data structures\n\n</div>\n\n### Poetry\n\nWant to know more about Poetry? Check [its documentation](https://python-poetry.org/docs/).\n\n<details>\n<summary>Details about Poetry</summary>\n<p>\n\nPoetry\'s [commands](https://python-poetry.org/docs/cli/#commands) are very intuitive and easy to learn, like:\n\n- `poetry add numpy@latest`\n- `poetry run pytest`\n- `poetry publish --build`\n\netc\n</p>\n</details>\n\n## Building and releasing your package\n\nBuilding a new version of the application contains steps:\n\n- Bump the version of your package `poetry version <version>`. You can pass the new version explicitly, or a rule such as `major`, `minor`, or `patch`. For more details, refer to the [Semantic Versions](https://semver.org/) standard.\n- Make a commit to `GitHub`.\n- Create a `GitHub release`.\n- And... publish 🙂 `poetry publish --build`\n\n\n## Installation\n\n```bash\npip install -U sparrow-datums\n```\n\nor install with `Poetry`\n\n```bash\npoetry add sparrow-datums\n```\n\nThen you can run\n\n```bash\nsparrow-datums --help\n```\n\nor with `Poetry`:\n\n```bash\npoetry run sparrow-datums --help\n```\n\n### Makefile usage\n\n[`Makefile`](https://github.com/sparrowml/sparrow-datums/blob/master/Makefile) contains a lot of functions for faster development.\n\n<details>\n<summary>1. Download and remove Poetry</summary>\n<p>\n\nTo download and install Poetry run:\n\n```bash\nmake poetry-download\n```\n\nTo uninstall\n\n```bash\nmake poetry-remove\n```\n\n</p>\n</details>\n\n<details>\n<summary>2. Install all dependencies and pre-commit hooks</summary>\n<p>\n\nInstall requirements:\n\n```bash\nmake install\n```\n\nPre-commit hooks coulb be installed after `git init` via\n\n```bash\nmake pre-commit-install\n```\n\n</p>\n</details>\n\n<details>\n<summary>3. Codestyle</summary>\n<p>\n\nAutomatic formatting uses `pyupgrade`, `isort` and `black`.\n\n```bash\nmake codestyle\n\n# or use synonym\nmake formatting\n```\n\nCodestyle checks only, without rewriting files:\n\n```bash\nmake check-codestyle\n```\n\n> Note: `check-codestyle` uses `isort`, `black` and `darglint` library\n\nUpdate all dev libraries to the latest version using one comand\n\n```bash\nmake update-dev-deps\n```\n\n<details>\n<summary>4. Code security</summary>\n<p>\n\n```bash\nmake check-safety\n```\n\nThis command launches `Poetry` integrity checks as well as identifies security issues with `Safety` and `Bandit`.\n\n```bash\nmake check-safety\n```\n\n</p>\n</details>\n\n</p>\n</details>\n\n<details>\n<summary>5. Type checks</summary>\n<p>\n\nRun `mypy` static type checker\n\n```bash\nmake mypy\n```\n\n</p>\n</details>\n\n<details>\n<summary>6. Tests with coverage badges</summary>\n<p>\n\nRun `pytest`\n\n```bash\nmake test\n```\n\n</p>\n</details>\n\n<details>\n<summary>7. All linters</summary>\n<p>\n\nOf course there is a command to ~~rule~~ run all linters in one:\n\n```bash\nmake lint\n```\n\nthe same as:\n\n```bash\nmake test && make check-codestyle && make mypy && make check-safety\n```\n\n</p>\n</details>\n\n<details>\n<summary>8. Docker</summary>\n<p>\n\n```bash\nmake docker-build\n```\n\nwhich is equivalent to:\n\n```bash\nmake docker-build VERSION=latest\n```\n\nRemove docker image with\n\n```bash\nmake docker-remove\n```\n\nMore information [about docker](https://github.com/sparrowml/sparrow-datums/tree/master/docker).\n\n</p>\n</details>\n\n<details>\n<summary>9. Cleanup</summary>\n<p>\nDelete pycache files\n\n```bash\nmake pycache-remove\n```\n\nRemove package build\n\n```bash\nmake build-remove\n```\n\nDelete .DS_STORE files\n\n```bash\nmake dsstore-remove\n```\n\nRemove .mypycache\n\n```bash\nmake mypycache-remove\n```\n\nOr to remove all above run:\n\n```bash\nmake cleanup\n```\n\n</p>\n</details>\n',
    'author': 'Sparrow Computing',
    'author_email': 'ben@sparrow.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sparrowml/sparrow-datums',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import json
import os
import tempfile

import numpy as np
import pytest

from sparrow_datums import BoxTracking

from .writer import ChunkStreamWriter


def test_constructor_writes_manifest_file():
    with tempfile.TemporaryDirectory() as dir:
        manifest_path = os.path.join(dir, "stream.jsonl")
        writer = ChunkStreamWriter(manifest_path, BoxTracking, fps=1)
        with open(manifest_path) as f:
            manifest_values = f.read().split("\n")
        assert len(manifest_values) == 2
        assert writer.footer.is_done == False
        writer.close()
        assert writer.footer.is_done


def test_writer_writes_chunks():
    with tempfile.TemporaryDirectory() as dir:
        manifest_path = os.path.join(dir, "stream.jsonl")
        writer = ChunkStreamWriter(manifest_path, BoxTracking, fps=1)
        a = BoxTracking(np.ones((2, 2, 4)), fps=1, start_time=0)
        b = BoxTracking(np.ones((2, 2, 4)), fps=1, start_time=a.duration)
        writer.add_chunk(a)
        writer.add_chunk(b)
        writer.close()
        with open(manifest_path) as f:
            manifest_values = f.read().split("\n")
        assert len(manifest_values) == 4
        for chunk_path in manifest_values[1:-1]:
            data = json.loads(chunk_path)
            _ = BoxTracking.from_file(os.path.join(dir, data["path"]))
        assert writer.footer.is_done


def test_context_manager_closes_stream():
    with pytest.raises(ValueError):
        with tempfile.TemporaryDirectory() as dir:
            manifest_path = os.path.join(dir, "stream.jsonl")
            with ChunkStreamWriter(manifest_path, BoxTracking, fps=1) as writer:
                assert writer.footer.is_done == False
                raise ValueError("foo bar")
    assert writer.footer.is_done


def test_first_chunk_can_set_header():
    with tempfile.TemporaryDirectory() as dir:
        manifest_path = os.path.join(dir, "stream.jsonl")
        writer = ChunkStreamWriter(manifest_path, BoxTracking)
        a = BoxTracking(np.ones((2, 2, 4)), fps=1, start_time=0)
        writer.add_chunk(a)
        assert writer.header.fps == 1
